/*
 *  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.  By
 *  downloading, copying, installing or using the software you agree to
 *  this license.  If you do not agree to this license, do not download,
 *  install, copy or use the software.
 *
 *  Intel Open Source License
 *
 *  Copyright (c) 2002 Intel Corporation
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *      Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *      Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *      Neither the name of the Intel Corporation nor the names of its
 *  contributors may be used to endorse or promote products derived from
 *  this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE INTEL OR ITS
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */
/*
 *
 * Authors: Phil Buonadonna,David Gay
 * Date last modified:  $Revision: 1.1 $
 *
 * Ported By: Junaith Ahemed
 * There is a change in the Exception Vector table to accomodate the
 * IMOTE2 Boot Loader. The branches point to a redirect function which
 * loads the address of Exception Handler function from RAM and jumps
 * to that location.
 * The reset vector always points to the boot loader Reset Exception
 * Handler.
 *
 * This file was initially taken from the original Bootloader in the 
 *  TinyOS repository.
 *
 * Modified by: Markus Engel, Networked Systems Group,
 *  University of Kaiserslautern
 * Changed the Exception Vector to be compatible with the exception vector of
 *  the later on running application, thus keeping debug capability using JTAG
 * Changed minor things such as SDRAM activation (SDRAM is also used for the
 *  text segment of the bootloader, such that it does not interfer with
 *  programming the flash) and added some comments
 */
 
.equ	ARM_CPSR_MODE_MASK,(0x0000001F)

.equ	ARM_CPSR_MODE_USR,(0x10)
.equ	ARM_CPSR_MODE_FIQ,(0x11)
.equ	ARM_CPSR_MODE_IRQ,(0x12)
.equ	ARM_CPSR_MODE_SVC,(0x13)
.equ	ARM_CPSR_MODE_ABT,(0x17)
.equ	ARM_CPSR_MODE_UND,(0x1B)
.equ	ARM_CPSR_MODE_SYS,(0x1F)
.equ 	VECTOR_ADDRS,(0x20)

.section .start

/* mark these routines as weak, thus they can be overridden when linking */
.weak hplarmv_undef, hplarmv_swi, hplarmv_pabort, hplarmv_dabort
.weak hplarmv_reserved, hplarmv_irq, hplarmv_fiq

/* exception vectors */
.section .vectors
.globl bootvectors
bootvectors:
	/* reset exception (RESET) */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	/* undefined instruction (UNDEF) */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	/* software interrupt (SWI) */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	/* prefetch (instruction) abort (PABORT) */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	/* data abort (DABORT) */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	/* reserved */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	/* interrupt request (IRQ) */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	/* fast interrupt request (FIQ) */
	ldr pc,[pc,#VECTOR_ADDRS-8]
	
.section .vecaddrs
.globl vectoraddresses
vectoraddresses:
	.long _reset_handler_start
	.long hplarmv_undef
	.long hplarmv_swi
	.long hplarmv_pabort
	.long hplarmv_dabort
	.long hplarmv_reserved
	.long hpl_irq_redir
	.long hpl_fiq_redir

/* vector table, located at 0x5c000000 in sram */
.section .vectable
irq_address:
	/* address of irq handler. The irq handler redirects to this location */
	.word hplarmv_irq
fiq_address:
	/* address of fiq handler. The fiq handler redirects to this location */
	.word hplarmv_fiq

/* macro copysection takes the load start, load end and the destination start.
	these are loaded into r0-r2.
	a loop loads 8*4 words (that are 128 bytes) from the load start and stores
		them to the destination, incrementing both start pointers
	the loop ends if the end is reached or exceeded
	in the worst case, up to 128 bytes are transfered after the end */
.macro copysection load_start load_end start
	ldr r0, =\load_start
	ldr r1, =\load_end
	ldr r2, =\start
1:
	ldmia r0!, {r4-r11}
	stmia r2!, {r4-r11}
	cmp r0, r1
	bls 1b
.endm


.section .start

.globl start
	/* start is some kind of trampoline function:
		it sets up the stacks and activates sdram.
		Then the text section is copied to sdram, the data section is
			copied to sram and the bss section (in sram) is cleared.
		At last, control is transfered to the copy of main in sdram */
start:
	/* the xscale processor has a bunch of operational modes (e.g.
		normal execution, supervisor (privileged), irq, fast irq,
		abort, etc) and each mode has its own stack pointer.
		
		For our purposes, the code runs in SVC (supervisor) mode, but it can
		switch into IRQ mode. Thus we'll set up these both stacks. */
	
	/* If any of the other modes (e.g. abort, undefined, fast interrupt) are
		used (and need stack space), the appropriate stack pointers MUST be
		initialized properly */
	
	/* the cpsr register controls the processor mode. When switching the mode,
		the stack register (SP) is switched, too */
	mrs r0, cpsr
	/* get only the modes of the cpsr register into r0 */
	bic r0, r0, #ARM_CPSR_MODE_MASK
	
	/* go to start of stack into r2 */
	ldr r2, =(__stack_start)
	
	/* first stack (at lowest address) is the svc stack */
	/* switch mode */
	mov r0, #ARM_CPSR_MODE_SVC
	msr cpsr_c, r0
	/* as stack grows down, go to the end of svc stack */
	ldr r1, =_SVC_STACK_SIZE
	add r2, r2, r1
	/* set svc stack pointer */
	mov sp, r2
	
	/* next is the irq stack */
	/* switch mode */
	mov r0, #ARM_CPSR_MODE_IRQ
	msr cpsr_c, r0
	/* go to end of irq stack */
	ldr r1, =_IRQ_STACK_SIZE
	add r2, r2, r1
	/* set irq stack pointer */
	mov sp, r2
	
	/* fiq stack: not used */
	mov r0, #ARM_CPSR_MODE_FIQ
	msr cpsr_c, r0
	mov sp, #0
	
	/* abt stack: not used */
	mov r0, #ARM_CPSR_MODE_ABT
	msr cpsr_c, r0
	mov sp, #0
	
	/* und stack: not used */
	mov r0, #ARM_CPSR_MODE_UND
	msr cpsr_c, r0
	mov sp, #0
	
	/* last of all, switch to svc mode */
	mov r0, #ARM_CPSR_MODE_SVC
	msr cpsr_c, r0
	
	/* stack setup done */
	
	/* copy data section */
	copysection __data_load_start, __data_load_end, __data_start
	
	/* enable sdram */
	bl sdramEnable
	
	/* copy text section */
	copysection __text_load_start, __text_load_end, __text_start
	
	
	/* clear bss section */
	ldr r0, =__bss_start__
	ldr r1, =__bss_end__
	mov r4, #0	/* move 0x0 into r4-r11 */
	mov r5, r4
	mov r6, r4
	mov r7, r4
	mov r8, r4
	mov r9, r4
	mov r10, r4
	mov r11, r4
1:
	stmia r0!, {r4-r11}	/* store r4-r11 into r0 and increment r0 */
	cmp r0, r1
	/* loop until end is reached or exceeded */
	blt 1b
	
	@bl _init
	
	mov r0, #0 /* argc */
	mov r1, #0 /* argv */
	mov r2, #0 /* envp */
	ldr lr, =1f	/* load address of next label in return register */
	ldr pc, =main /* long jump into main */
1:
	b 1b	/* endless loop */

/* default exception handlers. can be overridden because they are weak symbols */
hplarmv_undef:
	/* undefined instruction: indicate error condition */
	b error
hplarmv_swi:
	/* software interrupt: indicate error condition */
	b error
hplarmv_pabort:
	/* prefetch abort: indicate error condition */
	b error
hplarmv_dabort:
	/* data abort: indicate error condition */
	b error
hplarmv_reserved:
	/* reserved: indicate error condition */
	b error
hplarmv_irq:
	/* irq: indicate error condition (overridden by interrupt.c) */
	b error
hplarmv_fiq:
	/* fast irq: indicate error condition */
	/* XXX: if fast interrupts are ever used, you have to setup the
	 *  fiq stack properly in the start routine */
	b error

.section .start

.align 5

/* This is the reset handler routine. It determines wether it was called
 *  because of a debug event or a real reset and calls the debug handler or
 *  the start routine */
_reset_handler_start:
 /* NOTE: only r13 can be used safely in a debug exception */
 	/* get cpsr */
 	mrs r13, cpsr
 	/* apply mode mask */
 	and r13, r13, #ARM_CPSR_MODE_MASK
 	/* test if its the debug mode */
 	cmp r13, #0x15
 	/* if so, branch to the debug handler */
 	beq _dbg_handler_stub
 	/* if not, this was a normal reset.
 		enable debug and set MOE bits */
	mov r13, #0x8000001c
	mcr p14, 0, r13, c10, c0, 0
	/* continue by branching to the start routine which sets up stack and
		ram and finally starts the main routine */
	b start

.section .start

/* This routine is called whenever an interrupt occures, even if the image is
 *  already running.
 * Note also, that this method is copied into sdram by the start routine
 *  but executed from flash.
 * It redirects the irq to the appropriate irq handler.
 * The address of the irq handler is located at 0x5c000000 (=irq_address)
 *  by the bootloader and later replaced by the images irq handler address. */
.globl hpl_irq_redir
hpl_irq_redir:
	/* make space on stack for one field */
	sub sp, sp, #4
	/* save r0 on stack */
	stmfd sp!, {r0}
	/* load address of irq address */
	ldr r0, =irq_address
	/* load address */
	ldr r0, [r0]
	/* store in free slot on stack */
	str r0, [sp, #4]
	/* restore r0 from stack and jump into interrupt handler */
	ldmfd sp!, {r0, pc}

/* This routine is called whenever a fast interrupt occures, even if the image
 *  is already running.
 * Note also, that this method is copied into sdram by the start routine
 *  but executed from flash.
 * It redirects the irq to the appropriate fiq handler.
 *  This adress of the irq handler is located at 0x5c000004 (=fiq_address)
 * by the bootloader and later replaced by the images fiq handler address.
 * TODO because no Fast interrupts are used atm, this method will branch into
 *  error */
hpl_fiq_redir:
	/* TODO: this branches into error. If you want to use FIQs, setup that
	 * FIQ stack properly in the start routine and delete this line */
	b error
	
	/* fiq may use r8-r12 without saving them */
	/* load address of irq address */
	ldr r8, =fiq_address
	/* load address */
	ldr pc, [r8]

/* This routine is called whenever an error occures. This includes all unhandled
 *  exceptions.
 * It disables and invalidates all caches and tlbs and calls a routine which
 *  cycles the leds to indicate the error condition */
.globl error
error:
	/* restore stack pointer of previous mode */
	
	/* old mode in ip */
	mrs ip, SPSR
	/* current mode in r3 (still needed from code below) */
	mrs r0, CPSR
	
	/* keep interrupts disabled */
	orr ip, #0xc0
	
	/* restore old mode */
	msr	CPSR_c, ip
	
	/* save stack ptr */
	mov r1, sp
	
	/* restore current mode */
	msr CPSR, r0
	/* set this modes stack pointer */
	mov sp, r1
	
	/* unlock itlb */
	mcr p15, 0, r0, c10, c4, 1
	/* unlock dtlb */
	mcr p15, 0, r0, c10, c8, 1
	/* invalidate tlb */
	mcr p15, 0, r0, c8, c7, 0
	
	/* unlock icache */
	mcr p15, 0, r0, c9, c1, 1
	/* unlock dcache */
	mcr p15, 0, r0, c9, c2, 1
	/* invalidate caches and buffer */
	mcr p15, 0, r0, c7, c7, 0
	
	/* read control register */
	mrc p15, 0, r0, c1, c0, 0
	/* disable caches, believing the mmu still does it's work */
	ldr r1, =((1<<12)|(1<<11)|(1<<2))
	bic r0, r1
	mcr p15, 0, r0, c1, c0, 0
	
	/* cpwait */
	mrc p15, 0, r0, c2, c0, 0
	/* nop */
	mov r0, r0
	/* next instruction */
	sub pc, pc, #0x4
	
	/* cycle leds to indicate the error condition */
	b ledcycle

/* cycles the leds in an endless loop. used for indicating an error condition */
ledcycle:
	/* gpcr 3 (write to this register to switch a led on) */
	ldr r0, =0x40e00124
	/* gpsr 3 (write to this register to switch a led off) */
	ldr r1, =0x40e00118
	
	/* bit mask for clearing leds */
	ldr r2, =((1 << 7) | (1 << 8) | (1 << 9))
	
	/* first led in gpio register */
	mov r3, #(1<<7)
	/* number of leds */
	mov r4, #3
	
	/* put leds off */
	str r2, [r1]
	
	/* led status (all cleared) */
	mov r5, #0
	/* counter to do some busy waiting */
	mov r6, #(1<<23)
	
	/* first loop: busy wait, then set current led and increment led counter */
1:
	subs r6, #1
	bne 1b
	
	/* decrement led counter */
	subs r4, #1
	/* set current led (2 - 0) */
	orr r5, r5, r3, LSL r4
	/* reset led counter if its zero */
	moveq r4, #3
	/* put set leds on */
	str r5, [r0]
	/* reset busy waiting counter */
	mov r6, #(1<<23)

	/* second loop: busy wait, then clear current led and increment led counter */
2:
	subs r6, #1
	bne 2b
	
	/* decrement led counter */
	subs r4, #1
	/* clear current led (2-0) */
	bic r5, r5, r3, LSL r4
	/* reset led counter if its zero */
	moveq r4, #3
	/* invert leds and apply led mask */
	bic r7, r2, r5
	/* put unset leds off */
	str r7, [r1]
	/* reset busy waiting counter */
	mov r6, #(1<<23)
	
	/* jump into first loop again */
	b 1b

.section .start

/* align the code to begin on a cache line boundary */
.align 5
/* this is the debug handler */
_dbg_handler_stub:
	/* save the cache bits (CP 0) in lr */
	mrc p15, 0, r14, c1, c0, 0
	
	/* enable instruction cache */
	orr r13, r14, #0x1000
	mcr p15, 0, r13, c1, c0, 0
	
	/* save instruction cache enable bit in lr */
	and r13, #0x1000
	mov r14, r13, lsr #12
	
	/* sync operation */
	adr r13, line2
	mcr p15, 0, r13, c7, c5, 1
	
	/* invalidate btb */
	mcr p15, 0, r13, c7, c5, 6

	/* loop until TR is clear (see Developers Manual, Table 26-6 and
		the Core Manual 9.8.4) */
1:
	mrc p14, 0, r15, c14, c0, 0
	bvs 1b
	
	/* send the "ready for download" command */
	mov r13, #0x00b00000
	mcr p14, 0, r13, c8, c0, 0
	
	/* loop until command is sent completely */
1:
	mrc p14, 0, r15, c14, c0, 0
	bpl 1b
	
	/* restore the instruction cache configuration */
	mrc p15, 0, r13, c1, c0, 0
	tst r14, #0x1
	biceq r13, #0x1000
	mcr p15, 0, r13, c1, c0, 0
	
	/* read rx into r13 */
	mrc p14, 0, r13, c9, c0, 0
	/* jump there */
	mov pc, r13

.align 5
line2:
	.rept 8
	.word 0
	.endr
