/* Communication: Link between Communication and USB Driver
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <string.h>
#include "communication_buffer.h"
#include "communication_buffer_internal.h"
#include "../usb/usb_configuration.h"
#include "../usb/usb_request.h"
#include "../usb/usb_main.h"
#include "../debug.h"
#include "../dmalloc.h"
#include "../config.h"
#include "../common.h"

/** \defgroup communicationusb Communication Buffer Subsystem (USB)
 * \ingroup communication
 * @{
 */
 
/** \file communication_usb.c
 * \brief communication subsystem (usb): configuration and buffering
 */

/* callbacks */
static void start(struct CommunicationSystem * sys);
static void usbBusReset(const struct UsbDevice *devConf);
static void usbConfigured(const struct UsbDevice *devConf,
	uint8_t configIdx, uint8_t ifaceIdx, uint8_t settingIdx);
static void usbReceivedData(struct UsbRequest * req);
static void usbSendData(struct CommunicationSystem * sys,
	const void * data, uint32_t length);

/** \defgroup usbconfig USB Configuration
 * @{
 */

/** \brief Receiving Endpoint */
static struct UsbEndpoint usbEndpointRx = {
	.epAddress = 1,			/* endpoint address 1 */
	.epType = USB_ENDPOINT_TYPE_BULK,		/* bulk transfer */
	.epDirection = USB_ENDPOINT_DIR_OUT,	/* OUT endpoint (=recv) */
	.epMaxPacketSize = 64,	/* max packet size = 64 byte */
	
	.doubleBuffer = TRUE,
	.useDma = TRUE
};

/** \brief Transmitting Endpoint */
static struct UsbEndpoint usbEndpointTx = {
	.epAddress = 2,			/* endpoint address 2 */
	.epType = USB_ENDPOINT_TYPE_BULK,	/* bulk transfer */
	.epDirection = USB_ENDPOINT_DIR_IN,	/* IN endpoint (=send) */
	.epMaxPacketSize = 64,	/* max packet size = 64 byte */
	
	.doubleBuffer = TRUE,
	.useDma = TRUE
};

/** \brief Alternate Setting */
static struct UsbAltSetting usbAltSetting1 = {
	.altClass = 0xff,	/* vendor specific device class... */
	.altSubClass = 0x0,
	.altProtocol = 0x0,	/* ... and protocol */
	.altDescription = USB_COMM_INTERFACE_DESC,	/* description */
	
	.numEndpoints = 2,
	.endpoints = {
		&usbEndpointRx,
		&usbEndpointTx
	} /* end of endpoints */
};

/** \brief USB Interface */
static struct UsbInterface usbInterface1 = {
	.numAltSettings = 1,
	
	.altSettings = {
		&usbAltSetting1
	} /* end of alternate settings */
};

/** \brief USB Configuration */
static struct UsbConfiguration usbConfig1 = {
	.confSelfPowered = TRUE,	/* powered by bus */
	.confRemoteWakeup = FALSE,	/* no remote wakeups */
	.confMaxPower = USB_COMM_MAX_POWER,
	.confDescription = USB_COMM_CONFIG_DESC,

	.numInterfaces = 1,
	.interfaces = {
		&usbInterface1
	} /* end of interfaces */
};

/** \brief USB Device Configuration */
static struct UsbDevice usbDeviceConfig = {
	.devClass = 0,		/* defined at interface level */
	.devSubClass = 0,	/* see above */
	.devProtocol = 0,	/* ... */
	.devIdVendor = USB_COMM_ID_VENDOR,		/* vendor id */
	.devIdProduct = USB_COMM_ID_PRODUCT,	/* product id */
	.devBCDDevice = USB_COMM_DEVICE_VERSION,/* device version */
	.devManufacturer = USB_COMM_VENDOR_DESC,	/* manufacturer description */
	.devProduct = USB_COMM_PRODUCT_DESC,	/* product description */
	/*.devSerialNumber is filled by loadUsbConfig */
	
	.languageID = USB_COMM_LANGUAGE_ID,	/* language id */
	
	.cbBusReset = &usbBusReset,		/* callback on bus reset */
	.cbConfigured = &usbConfigured,	/* callback on configure request */

	.numConfigurations = 1,
	.configurations = {
		&usbConfig1
	}
};

/** @} */

struct CommunicationSystem usbCommunicationSystem = {
	.start = &start,
	
	.send = &usbSendData
};

/** \brief start usb communication
 *
 * Loads the usb configuration */
static void start(struct CommunicationSystem * sys) {
	static char buf[20]; /* must be static as it is used outside the function */
	
	/* serial number of the device is located in flash */
	sprintf(buf, "%08x:%u", *(const uint32_t*)COMM_SERIAL_NUMBER_LOC,
		*(const uint32_t*)COMM_NODE_ID_LOC);
	usbDeviceConfig.devSerialNumber = buf;
	
	/* load configuration */
	usbConfigLoad(&usbDeviceConfig);
}

/** \brief Callback for USB bus reset
 * \param devConf the device configuration
 * \note called from <b>Interrupt context</b> */
static void usbBusReset(const struct UsbDevice *devConf) {
	communicationReset();
}

/** \brief Callback for configured USB bus
 * \param devConf the device configuration
 * \param configIdx current configuration index
 * \param ifaceIdx current interface index
 * \param settingIdx current alternate setting index
 * \note called from <b>Interrupt context</b> */
static void usbConfigured(const struct UsbDevice *devConf,
	uint8_t configIdx, uint8_t ifaceIdx, uint8_t settingIdx) {
	static struct UsbRequest rxReq;
	static aligned32 uint8_t rxData[COMM_PKT_SIZE];
	
	communicationConfigured();
	
	usbRequestFill(&rxReq, &usbEndpointRx, rxData, sizeof(rxData),
		&usbReceivedData);
	usbRequestEnqueue(&rxReq);
}

static void usbSendData(struct CommunicationSystem * sys,
	const void * data, uint32_t length) {
	static struct UsbRequest txReq;
	
	usbRequestFill(&txReq, &usbEndpointTx, (void*)data, length, NULL);
	usbRequestEnqueue(&txReq);
}


/** \brief Callback when data received on \b OUT endpoint
 *
 * buffer it and pass to communiation system if packet is fully received
 *
 * \param req usb request
 * \note called from <b>Interrupt context</b> */
static void usbReceivedData(struct UsbRequest * req) {
	if (req->status == USB_REQUEST_STATUS_SUCCESS) {
		communicationReceive(req->data, req->actual);
		usbRequestEnqueue(req);
	}
}

/** @} */

