/* Core Hardware: Cache management
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

.include "core_internal.s"

/** \brief enable the instruction cache
 * \details unlocks and invalidates the instruction cache and instruction tlb
 *  and activates the instruction cache
 * \note as a side-effect, the branch target buffer is also invalidated
 * \note look at the Intel XScale Core Manual on Page 4.3.2-52 (ICache) and
 *  7.2.2-83 (Control Register) */
.globl iCacheEnable
iCacheEnable:
.func iCacheEnable
	/* unlock instruction cache */
	ICACHE_UNLOCK
	/* unlock instruction tlb */
	ITLB_UNLOCK
	/* invalidate instruction cache */
	ICACHE_INVALIDATE
	/* invalidate instruction tlb */
	ITLB_INVALIDATE
	/* set I-Bit to activate cache */
	SET_CONTROL_BITS r0, #CP15_CONTROL_I
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief determine wether the instruction cache is activated
 * \return \c TRUE iff cache is enabled
 * \note look at the Intel XScale Core Manual on
 *  Page 7.2.2-83 (Control Register) */
.globl iCacheIsEnabled
iCacheIsEnabled:
.func iCacheIsEnabled
	/* test I-Bit and save result (true or false) in r0 */
	TEST_CONTROL_BITS r0, #CP15_CONTROL_I
	/* return */
	mov pc, lr
.endfunc

/** \brief disable the instruction cache
 * \details disables the instruction cache, unlocks and invalidates the cache,
 *  branch target buffer and tlb
 * \note look at the Intel XScale Core Manual on Page 4.3.2-52 (ICache)
 *  and 7.2.2-83 (Control Register) */
.globl iCacheDisable
iCacheDisable:
.func iCacheDisable
	/* unlock instruction cache */
	ICACHE_UNLOCK
	/* unlock instruction tlb */
	ITLB_UNLOCK
	/* unset I-Bit to deactivate the cache */
	UNSET_CONTROL_BITS r0, #CP15_CONTROL_I
	/* wait for changes */
	CPWAIT r0
	/* invalidate instruction cache */
	ICACHE_INVALIDATE
	/* invalidate instruction tlb */
	ITLB_INVALIDATE
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief invalidate the instruction cache
 * \note as a side effect, the branch target buffer is also invalidated
 * \note this will not invalidate locked regions. If you need this, call
 *  iCacheUnlock() before
 * \note look at the XScale Core Manual on Page 4.3.3-53 (ICache) and
 *  7.2.8-78 (Cache Functions) */
.globl iCacheInvalidate
iCacheInvalidate:
.func iCacheInvalidate
	/* invalidate the instruction cache and btb */
	ICACHE_INVALIDATE
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief invalidate consecutive lines of the instruction cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \note This will not invalidate the branch target buffer. If you need this,
 *  call btbInvalidate() afterwards
 * \warning the given line should not be locked as this results in an unusable
 *  cache line
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
.globl iCacheInvalidateLine
iCacheInvalidateLine:
.func iCacheInvalidateLine
	/* test size against zero */
	cmp r1, #0
	/* do nothing if size is zero */
	moveq pc, lr
	
	/* align base address */
	bic r0, r0, #31
1:
	/* invalidate instruction cache line denoted by r0 */
	mcr p15, 0, r0, c7, c5, 1
	/* add size of a line for next address */
	add r0, r0, #32
	/* subtract size of a line */
	subs r1, r1, #32
	/* loop if not done (r1 > 0) */
	bgt 1b
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief lock instructions in the instruction cache
 * \param addr base address
 * \param length length
 * \note look at the XScale Core Manual on Page 4.3.4-54 (ICache) and
 *  7.2.10-90 (Cache Lock Down)
 * \note this code must be placed in a non-cacheable region as described in the
 *  Core Manual */
/*.globl iCacheLock
iCacheLock:
.func iCacheLock
	@ not implemented yet
.endfunc*/

/** \brief unlock the instruction cache
 * \note look at the XScale Core Manual on Page 7.2.10-90 (Cache Lock Down) */
.globl iCacheUnlock
iCacheUnlock:
.func iCacheUnlock
	/* unlock the instruction cache */
	mcr p15, 0, r0, c9, c1, 1
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief enable the branch target buffer
 * \details invalidates the buffer and activates it
 * \note look at the XScale Core Manual on Page 5.2.1-59 (BTB) and
 *  7.2.2-83 (Control Register) */
.globl btbEnable
btbEnable:
.func btbEnable
	/* invalidate btb */
	BTB_INVALIDATE
	/* wait for changes */
	CPWAIT r0
	/* set Z-Bit to activate the branch target buffer */
	SET_CONTROL_BITS r0, #CP15_CONTROL_Z
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief determine wether the branch target buffer is enabled
 * \return \c TRUE iff the buffer is enabled
 * \note look at the XScale Core Manual on Page 7.2.2-83 (Control Register) */
.globl btbIsEnabled
btbIsEnabled:
.func btbIsEnabled
	/* test Z-Bit and save result (true or false) in r0 */
	TEST_CONTROL_BITS r0, #CP15_CONTROL_Z
	/* return */
	mov pc, lr
.endfunc

/** \brief disable the branch target buffer
 * \details invalidates the buffer and deactivates it
 * \note look at the XScale Core Manual on Page 5.2.1-59 (BTB) and
 *  7.2.2-83 (Control Register) */
.globl btbDisable
btbDisable:
.func btbDisable
	/* invalidate btb */
	BTB_INVALIDATE
	/* wait for changes */
	CPWAIT r0
	/* unset Z-Bit to deactivate the branch target buffer */
	UNSET_CONTROL_BITS r0, #CP15_CONTROL_Z
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief invalidate the branch target buffer
 * \note look at the XScale Core Manual on Page 5.2.2-59 (BTB) and
 *  7.2.8-78 (Cache Functions) */
.globl btbInvalidate
btbInvalidate:
.func btbInvalidate
	/* invalidate btb */
	BTB_INVALIDATE
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief enable the data cache.
 * \details unlocks and invalidates the data cache and data tlb,
 *  flushes the write buffer and activates the cache
 * \warning make sure that the MMU is enabled at this point
 * \note look at the XScale Core Manual on Page 6.3.2-69 (DCache) and
 *  7.2.2-83 (Control Register) */
.globl dCacheEnable
dCacheEnable:
.func dCacheEnable
	/* unlock data cache */
	DCACHE_UNLOCK
	/* unlock data tlb */
	DTLB_UNLOCK
	/* invalidate data cache */
	DCACHE_INVALIDATE
	/* invalidate data tlb */
	DTLB_INVALIDATE
	/* drain write buffer */
	DCACHE_DRAIN_BUFFER
	/* set C-Bit to activate cache */
	SET_CONTROL_BITS r0, #CP15_CONTROL_C
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief determine wether the data cache is activated
 * \return \c TRUE iff cache is enabled
 * \note look at the Intel XScale Core Manual on
 *  Page 7.2.2-83 (Control Register) */
.globl dCacheIsEnabled
dCacheIsEnabled:
.func dCacheIsEnabled
	/* test C-Bit and save result (true or false) in r0 */
	TEST_CONTROL_BITS r0, #CP15_CONTROL_C
	/* return */
	mov pc, lr
.endfunc

/** \brief disable the data cache
 * \details flushes the write buffer, disables the data cache, unlocks and
 *  invalidates the cache and tlb
 * \note look at the Intel XScale Core Manual on Page 6-61 (DCache),
 *  7.2.2-83 (Control Register) */
.globl dCacheDisable
dCacheDisable:
.func dCacheDisable
	/* save lr */
	stmfd sp!, {lr}
	/* drain write buffer */
	DCACHE_DRAIN_BUFFER
	/* unset C-Bit do deactivate the cache */
	UNSET_CONTROL_BITS r0, #CP15_CONTROL_C
	/* unlock data cache */
	DCACHE_UNLOCK
	/* unlock data tlb */
	DTLB_UNLOCK
	/* clean and invalidate data cache */
	bl dCacheCleanAndInvalidate
	/* invalidate data tlb */
	DTLB_INVALIDATE
	/* wait for changes */
	CPWAIT r0
	/* return by restoring lr into pc */
	ldmfd sp!, {lr}
.endfunc

/** \brief clean and invalidate the data and mini-data cache
 * \details cleans the data cache by allocating lines in it, then cleans the
 *  mini-data cache by loading abitrary data in it and then invalidates the
 *  caches
 * \note look at the Intel XScale Core Manual on Page 6.3.3.1-70 (DCache) */
.globl dCacheCleanAndInvalidate
dCacheCleanAndInvalidate:
.func dCacheCleanAndInvalidate
	/* cacheable but non-existent memory for cleaning cache */
	ldr r1, =0x04000000
	/* add number of cache lines multiplied by line size */
	add r0, r1, #(1024 * 32)
1:
	/* allocate line */
	DCACHE_ALLOCATE r1
	/* advance to next line */
	add r1, r1, #32
	/* compare begin against end */
	cmp r0, r1
	/* loop if not done */
	bne 1b
	
	/* r1 = 0x0400800, cacheable but non-existent memory for cleaning cache */
	/* add number of cache lines multiplied by line size */
	add r0, r1, #(64 * 32)
1:
	/* read data and advance to next line */
	ldr r2, [r1], #32
	/* compare begin against end */
	cmp r0, r1
	/* loop if not done */
	bne 1b
	
	/* invalidate data cache */
	DCACHE_INVALIDATE
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief invalidate consecutive lines of the data cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \warning This will not clean the cache. If you need this, call
 *  dCacheCleanLine() before
 * \warning the given line should not be locked as this results in an unusable
 *  cache line
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
.globl dCacheInvalidateLine
dCacheInvalidateLine:
.func dCacheInvalidateLine
	/* test size against zero */
	cmp r1, #0
	/* do nothing if size is zero */
	moveq pc, lr
	
	/* align base address */
	bic r0, r0, #31
1:
	/* invalidate data cache line denoted by r0 */
	DCACHE_INVALIDATE_LINE r0
	/* add size of a line for next address */
	add r0, r0, #32
	/* subtract size of a line */
	subs r1, r1, #32
	/* loop if not done (r1 > 0) */
	bgt 1b
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief clean consecutive lines of the data cache
 * \param addr base address to be cleaned
 * \param len length of the area to be cleaned
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
.globl dCacheCleanLine
dCacheCleanLine:
.func dCacheCleanLine
	/* test size against zero */
	cmp r1, #0
	/* do nothing if size is zero */
	moveq pc, lr
	
	/* align base address */
	bic r0, r0, #31
1:
	/* clean data cache line denoted by r0 */
	DCACHE_CLEAN_LINE r0
	/* add size of a line for next address */
	add r0, r0, #32
	/* subtract size of a line */
	subs r1, r1, #32
	/* loop if not done (r1 > 0) */
	bgt 1b
	/* drain write buffer */
	DCACHE_DRAIN_BUFFER
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief clean and invalidate consecutive lines of the data cache
 * \param addr base address to be cleaned and invalidated
 * \param len length of the area to be cleaned and invalidated
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
.globl dCacheCleanAndInvalidateLine
dCacheCleanAndInvalidateLine:
.func dCacheCleanAndInvalidateLine
	/* test size against zero */
	cmp r1, #0
	/* do nothing if size is zero */
	moveq pc, lr
	
	/* align base address */
	bic r0, r0, #31
1:
	/* clean data cache line denoted by r0 */
	DCACHE_CLEAN_LINE r0
	/* invalidate data cache line denoted by r0 */
	DCACHE_INVALIDATE_LINE r0
	/* add size of a line for next address */
	add r0, r0, #32
	/* subtract size of a line */
	subs r1, r1, #32
	/* loop if not done (r1 > 0) */
	bgt 1b
	/* drain write buffer */
	DCACHE_DRAIN_BUFFER
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief drain write buffer
 * \note look at the XScale Core Manual on Page 6.1.3-64 (DCache)
 *  and 7.2.8-78 (Cache Functions) */
.globl dCacheDrainBuffer
dCacheDrainBuffer:
.func dCacheDrainBuffer
	/* flush buffer */
	DCACHE_DRAIN_BUFFER
	/* wait for changes */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

