/* CFI Flash driver
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef FLASH_INTERNAL_H
#define FLASH_INTERNAL_H

#include "config.h"

/** \cond INTERNAL */

/** \addtogroup flash
 * @{
 */

/** \file flash_internal.h
 * \brief internal flash structures and definitions
 */

/** \brief ident offset */
#define OFFSET_CFI_IDENT 0x10

/** \brief command <a>READ ARRAY</a> */
#define CMD_READ_ARRAY		0xFF
/** \brief command <a>READ QUERY</a> */
#define CMD_READ_QUERY		0x90
/** \brief command <a>READ STATUS</a> */
#define	CMD_READ_STATUS		0x70
/** \brief command <a>CLEAR STATUS</a> */
#define	CMD_CLEAR_STATUS	0x50

/** \brief command <a>BLOCK ERASE</a> */
#define CMD_BLOCK_ERASE		0x20
/** \brief command <a>LOCK BLOCK</a> */
#define CMD_LOCK_BLOCK		0x60

/** \brief command <a>ERASE CONFIRM</a> */
#define CMD_ERASE_CONFIRM	0xD0

/** \brief command <a>PROGRAM WORD</a> */
#define	CMD_WORD_PROGRAM	0x40
/** \brief command <a>WRITE TO BUFFER</a> */
#define CMD_WRITE_TO_BUFFER_SETUP 0xE8
/** \brief command <a>WRITE TO BUFFER CONFIRM</a> */
#define CMD_WRITE_TO_BUFFER_CONFIRM 0xD0

/** \brief status bit <a>READY</a> */
#define STATUS_READY	(1 << 7)
/** \brief status bit <a>ERASE ERROR</a> */
#define STATUS_ERASE_ERR (1 << 5)
/** \brief status bit <a>PROGRAM ERROR</a> */
#define STATUS_PGM_ERR	(1 << 4)

/** \brief lock bit <a>LOCK</a> */
#define BLOCK_LOCK			0x01
/** \brief lock bit <a>UNLOCK</a> */
#define BLOCK_UNLOCK		0xD0
/** \brief lock bit <a>LOCKDOWN</a> */
#define BLOCK_LOCKDOWN		0x2F

/** \brief status lock bit <a>LOCKED</a> */
#define LOCK_STATUS_LOCKED (1 << 0)
/** \brief status lock bit <a>LOCKDOWN</a> */
#define LOCK_STATUS_LOCKEDDOWN (1 << 1)

/** \brief calculate pointer to 16 byte location from base and offset
 * \param block block (added to base)
 * \param offset offset in block (doubled and added to block)
 * \return absolute address to location */
#define ADDR16(block, offset) \
	(((volatile uint16_t*)((FLASH_BASE_ADDR) + (block))) + (offset))

/** \brief Basic Query Structure */
struct CfiIdent {
	/** \brief query identifier, must be 'Q' 'R' 'Y' */
	uint8_t qry[3];
	/** \brief primary command set */
	uint16_t cmdSet;
	/** \brief address to primary extension table */
	uint16_t extTable;
	/** \brief alternate command set */
	uint16_t altCmdSet;
	/** \brief address to alternate extension table */
	uint16_t altExtTable;
	/** \brief VCC min (4.4 bit) */
	uint8_t vccMin;
	/** \brief VCC max (4.4 bit) */
	uint8_t vccMax;
	/** \brief VPP min (4.4 bit) */
	uint8_t vppMin;
	/** \brief VPP max (4.4 bit) */
	uint8_t vppMax;
	/** \brief typical word write timeout in ld(x) us */
	uint8_t wordWriteTimeoutTyp;
	/** \brief typical buffer write timeout in ld(x) us */
	uint8_t bufWriteTimeoutTyp;
	/** \brief typical block erase timeout in ld(x) ms */
	uint8_t blockEraseTimeoutTyp;
	/** \brief typical chip erase timeout in ld(x) ms */
	uint8_t chipEraseTimeoutTyp;
	/** \brief max word write timeout in ld(x) / ld(typ) us */
	uint8_t wordWriteTimeoutMax;
	/** \brief max buffer write timeout in ld(x) / ld(typ) us */
	uint8_t bufWriteTimeoutMax;
	/** \brief typical block erase timeout in ld(x) / ld(typ) ms */
	uint8_t blockEraseTimeoutMax;
	/** \brief typical chip erase timeout in ld(x) / ld(typ) ms */
	uint8_t chipEraseTimeoutMax;
	/** \brief device size in ld(x) bytes */
	uint8_t devSize;
	/** \brief interface description */
	uint16_t interfaceDesc;
	/** \brief max buffer write size in ld(x) bytes */
	uint16_t maxBufWriteSize;
	/** \brief number of erase regions */
	uint8_t numEraseRegions;
	/** \brief erase regions */
	uint32_t eraseRegionInfo[0];
} __attribute__((packed));

/** \brief PRI for Intel Command Set */
struct CfiPriIntelExt {
	/** \brief query identifier, must be 'P' 'R' 'I' */
	uint8_t pri[3];
	/** \brief major version as ascii character */
	uint8_t majorVersion;
	/** \brief minor version as ascii character */
	uint8_t minorVersion;
	/** \brief bitset for features */
	uint32_t featureSupport;
	/** \brief bitset for suspend cmd support */
	uint8_t suspendCmdSupport;
	/** \brief block status register mask */
	uint16_t blkStatusRegMask;
	/** \brief VCC optimum (4.4 bytes) */
	uint8_t vccOptimal;
	/** \brief VPP optimum (4.4 bytes) */
	uint8_t vppOptimal;
	/** \note not used by now */
	uint16_t protRegAddr;
	/** \note not used by now */
	uint8_t factProtRegSize;
	/** \note not used by now */
	uint8_t userProtRegSize;
} __attribute__((packed));

/** @} */

/** \endcond */

#endif

