/* Image definitions, detection and booting
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "image.h"
#include "debug.h"
#include "config.h"
#include "utils.h"

/** \file image.c
 * \brief All image related stuff, like validity checking, retrieving header
 *  information and finally booting
 */

/** Image header */
struct ImageHeader {
	/** header signature, must be 0xcafebabe */
	uint32_t magic;
	/** header version */
	uint32_t version;
	
	/** image size */
	uint32_t size;
	
	/** image virtual memory address */
	const void * vma;
	/** image entry (w.r.t. vma) */
	const void * entry;
	/** image description (w.r.t. vma) */
	const void * description;
} __attribute__((packed));

/** header size */
#define IMG_HEADER_SIZE (8 * sizeof(uint32_t))
/** header version */
#define IMG_HEADER_VERSION (0x0001000)
/** header offset */
#define IMG_HEADER 0x40
/** header magic */
#define IMG_MAGIC 0xcafebabe
/** image end magic */
#define IMG_END_MAGIC 0xdeadc0de

/* compat */

/** Old Image Header */
struct ImageOldHeader {
	/** header signature, must be 0xcafebabe */
	uint32_t magic;
	/** image size - 4*/
	uint32_t size;
} __attribute__((packed));

/** old image header offset */
#define IMG_OLD_HEADER 0x60
/** old images entry */
#define IMG_OLD_ENTRY 0x20

static void * relocate(const void * addr, const struct Image * img) {
	return (void*)((uint32_t)addr - (uint32_t)img->vma +
		(uint32_t)img->lma);
}

void imageReadDefault(struct Image * img) {
	imageRead((const void*)IMG_DEFAULT_LOC, FLASH_SIZE - IMG_DEFAULT_LOC, img);
}

static bool tryReadNewImage(const void * lma, uint32_t maxaddr,
	struct Image * img);
static bool tryReadOldImage(const void * lma, uint32_t maxaddr,
	struct Image * img);

void imageRead(const void * lma, size_t maxsize, struct Image * img) {
	uint32_t maxaddr = (uint32_t)lma + maxsize;
	
	debugf("Reading image at location 0x%08x with maxsize %u(=0x%x)",
		(uint32_t)lma, (unsigned int)maxsize, (unsigned int)maxsize);
	
	/* prepare image structure */
	img->lma = lma;
	
	img->size = 0;
	img->version = 0;
	img->description = NULL;
	
	img->bootable = FALSE;
	
	if (tryReadNewImage(lma, maxaddr, img)) {
		img->bootable = TRUE;
		img->versionFlag = IMAGE_VERSION_NEW;
	} else if (tryReadOldImage(lma, maxaddr, img)) {
		img->bootable = TRUE;
		img->versionFlag = IMAGE_VERSION_OLD;
	}
}

static bool tryReadNewImage(const void * lma, uint32_t maxaddr,
	struct Image * img) {
	const struct ImageHeader * hdr =
		(const struct ImageHeader*)((uint32_t)lma + IMG_HEADER);
	
	debug("Checking wether image is a new image");
	
	if (((uint32_t)hdr) + IMG_HEADER_SIZE > maxaddr) {
		/* cannot read header, because maxsize is smaller,
		 * thus it is not a new image */
		return FALSE;
	}
	
	if (hdr->magic == 0xffffffff &&
		*(const uint32_t*)lma == 0xffffffff) {
		/* location seems to be unprogrammed */
		debug("\tImage location is unprogrammed");
		return FALSE;
	}
	
	if (hdr->magic != IMG_MAGIC) {
		/* magic1 check did not pass => not a new image */
		debug("\tImage did not pass magic check, not a new image");
		return FALSE;
	}
	
	const uint32_t * endptr = (const uint32_t*)((uint32_t)lma + hdr->size);
	if ((uint32_t)endptr > maxaddr) {
		/* image is too big (which is not really possible) or
		 * size field is corrupted */
		debug("\tImage size could not be read, assuming corrupted image");
		return FALSE;
	}
	
	if (endptr[-1] != IMG_END_MAGIC) {
		/* image is corrupted */
		debug("\tImage end magic did not pass, assuming corrupted image");
		return FALSE;
	}
	
	/* read header fields into image info */
	img->size = hdr->size;
	img->vma = hdr->vma;
	img->entry = (void(*)())relocate(hdr->entry, img);
	img->version = hdr->version;
	img->description = (const char *)relocate(hdr->description, img);
	
	debugf("\tImage check detected a new image: Entry 0x%08x, Version %d.%d, "
		"Description \"%s\"",
		(uint32_t)img->entry, img->version >> 16, img->version & 0xffff,
		img->description);
	
	return TRUE;
}

static bool tryReadOldImage(const void * lma, uint32_t maxaddr,
	struct Image * img) {
	const struct ImageOldHeader * hdr =
		(const struct ImageOldHeader*)((uint32_t)lma + IMG_OLD_HEADER);
	
	debug("Checking wether image is an old image");
	
	if (((uint32_t)hdr) + sizeof(*hdr) > maxaddr) {
		/* cannot read header, because maxsize is smaller,
		 * thus it is not an old image */
		return FALSE;
	}
	
	if (hdr->magic == 0xffffffff &&
		*(const uint32_t*)lma == 0xffffffff) {
		/* location seems to be unprogrammed */
		debug("\tImage location is unprogrammed");
		return FALSE;
	}
	
	if (hdr->magic != IMG_MAGIC) {
		/* magic1 check did not pass => not an old image */
		debug("\tImage did not pass magic check, not an old image");
		return FALSE;
	}
	
	const uint32_t * endptr = (const uint32_t*)((uint32_t)lma + hdr->size);
	if ((uint32_t)endptr - 4 > maxaddr) {
		/* image is too big (which is not really possible) or
		 * size field is corrupted */
		debug("\tImage size could not be read, assuming corrupted image");
		return FALSE;
	}
	
	if (*endptr != IMG_END_MAGIC) {
		/* image is corrupted */
		debug("\tImage end magic did not pass, assuming corrupted image");
		return FALSE;
	}
	
	/* read header fields into image info */
	img->size = hdr->size + 4;
	img->vma = lma;
	img->entry = (void(*)())((uint32_t)lma + IMG_OLD_ENTRY);
	
	debug("\tImage check detected an old image");
	
	return TRUE;
}

__attribute__((noreturn))
void imageBoot(const struct Image * img) {
	if (!img->bootable) {
		debug("Image is not bootable");
		error();
	}
	bootTo(img->entry);
}

