/* Main Routine
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <os_arm_bare_registers_def.h>
#include <os_arm_bare.h>
#include <os_arm_bare_interrupt.h>
#include <os_arm_bare_inithardware.h>
#include "interrupt.h"
#include "usb/usb_main.h"
#include "dma.h"
#include "debug.h"
#include "led.h"
#include "communication/communication_buffer.h"
#include "communication/communication_usb.h"
#include "task.h"
#include "flash.h"
#include "image.h"
#include "utils.h"
#include "config.h"
#include "timer.h"
#include "deploy.h"
#include "driver.h"
#include "core/cache.h"
#include "usb/usb_main_internal.h"
#include "usb/usb_request.h"

void mmuTableLoad();

/** \file main.c
 * \brief Main file
 * 
 * Includes the main function which initializates all drivers and
 *  runs a task list which is filled by the drivers */

static struct Image image;

__attribute__((noreturn))
static void bootTimeout(uint8_t timerId);

/** \brief main routine: initialize hardware, check for usb, run bottom halves
 *  and boot.
 * \note \a noreturn as routine does never return
 * \note \a no_instrument_function as the debug interface is not initialized
 *  at this point, so disable tracing for that routine */
__attribute__((noreturn,no_instrument_function))
int main() {
#ifdef DEBUG
	/* initialize debug interface */
	driverStart(&debugDriver);
#endif
	
	/* initialize & enable some core hardware */
	mmuTableLoad();
	dCacheEnable();
	iCacheEnable();
	btbEnable();
	
	/* initialize hardware (cpu frequency, interrupts) */
	hardwareInit();
	
	/* initialize interrupts */
	driverStart(&interruptDriver);
	
	/* initialize timers */
	driverStart(&timerDriver);
	
	/* initialize and start DMA */
	driverStart(&dmaDriver);
	
	/* setup PSSR as defined on page 3-72 of the developers manual */
	PSSR = PSSR_RDH | PSSR_PH;
	
	driverInit(&ledDriver);
	
	driverInit(&usbDriver);
	
	/* setup PSSR (3-72) again */
	PSSR = 0;
	
	/* initialize flash interface */
	driverStart(&flashDriver);
	
	deploy();
	
	/* validate image */
	imageReadDefault(&image);
	
	utilFlashLedStart();
	
	if (!image.bootable) {
		/* image is not valid, light up red led and wait for new image */
		debug("Image is not valid, waiting for a new image");
		ledOn(LED_RED);
		
		/* start communication system */
		communicationStart(&usbCommunicationSystem);
		
		/* start usb subsystem */
		driverStart(&usbDriver);
	} else if (usbBusPowered()) {
		/* usb connected */
		debug("Powered by USB");
		
		/* start boot timeout */
		utilStartBootTimeout(&bootTimeout);
		
		/* start communication system */
		communicationStart(&usbCommunicationSystem);
		
		/* start usb subsystem */
		driverStart(&usbDriver);
	} else {
		/* usb unconnected */
		debug("Powered by Battery");
		
		/* boot, this routine will not return */
		imageBoot(&image);
	}
	
	while (TRUE) {
		/* run pending tasks if there are any */
		taskRun();
#if defined(POWER_SAVE) && (POWER_SAVE == TRUE)
		SENF_DISABLE_INTERRUPTS();
		if (!taskPending()) {
			setCR7_Register(PWRMODE_M_IDLE);
		}
		SENF_ENABLE_INTERRUPTS();
#endif
	}
}

/** \brief boot timer fired
 *
 * boot timer fired, stop usb and boot
 * \param timerId timer which raised the event
 * \note \a noreturn as routine does never return
 * \note called from <b>Interrupt context</b> */
__attribute__((noreturn))
static void bootTimeout(uint8_t timerId) {
	debug("Boot timeout");
	/* boot */
	imageBoot(&image);
}

