/* USB Driver: Control Pipe Transfer and State Machine
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <os_arm_bare_registers_def.h>
#include <inttypes.h>
#include <string.h>
#include "usb_descriptors.h"
#include "usb_configuration.h"
#include "usb_control_pipe_internal.h"
#include "usb_main_internal.h"
#include "usb_request.h"
#include "usb_request_internal.h"
#include "usb_endpoint_internal.h"
#include "usb_debug.h"
#include "../dmalloc.h"
#include "../assert.h"
#include "../common.h"

/** \cond INTERNAL */

/** \defgroup usbcontrol USB Control Pipe (Endpoint zero)
 * \ingroup usbcomm
 * @{
 */
 
/** \file usb_control_pipe.c
 * \brief USB Control Pipe
 *
 * handling packets on endpoint zero
 */

static void readAndHandleControlRequest();
static bool handleControlRequest(const struct DeviceRequest * req);
static void sendDeviceDescriptor(uint16_t wLength);
static void sendConfigDescriptor(uint8_t index, uint16_t wLength);
static void sendStringDescriptor(uint8_t index, uint16_t wLength);
static void sendDescriptor(const void * descriptor, uint8_t size,
	uint16_t maxLength, bool freeData);
static void cbFreeRequest(struct UsbRequest * req);

/** \brief usb request to send descriptors over endpoint 0 */
struct UsbRequest usbRequest0;

/** \brief endpoint zero configuration */
struct UsbEndpoint usbEndpoint0 = {
	.epAddress = 0,
	.epType = USB_ENDPOINT_TYPE_CONTROL,
	.epDirection = USB_ENDPOINT_DIR_IN,
	.epMaxPacketSize = 16,
	.doubleBuffer = FALSE,
	.useDma = FALSE,
	
	.udcEndpoint = UDC_EP_0,
	/* dma stuff is ignored for endpoint zero */
	.enabled = TRUE,
	
	.req = NULL
};

/** \brief initialize control pipe */
void usbControlPipeInit() {
	usbEndpoint0.enabled = TRUE;
	usbEndpoint0.req = NULL;
}

/** \brief set control pipe state
 * \param cpState new control pipe state */
void usbControlPipeSetState(enum USB_CONTROL_PIPE_STATE cpState) {
#if DEBUG_USB_CP_LEVEL >= 3
	static const char * NAMES[] = {
		"WAIT_FOR_SETUP",
		"SETUP_STAGE",
		"IN_DATA_STAGE",
		"OUT_DATA_STAGE",
		"IN_STATUS_STAGE",
		"OUT_STATUS_STAGE",
		"STALL" };
	enum USB_CONTROL_PIPE_STATE old = usbUdc.cpState;
#endif
	usbUdc.cpState = cpState;
	usbCPDebug("Setting state from %s to %s", NAMES[old], NAMES[cpState]);
}

/** \brief control pipe interrupt handling
 * \param pcInt true iff packet complete interrupt occured
 * \param feInt true iff fifo error interrupt occured */
void usbControlPipeInterrupt(bool pcInt, bool feInt) {
	uint32_t udccsr0 = UDCCSR0;
	
	struct UsbRequest * req = usbEndpoint0.req;
	
	/* handle stalls */
	if (udccsr0 & UDCCSR0_SST) {
		usbCPErr("Clearing stall");
		usbEndpointNuke(&usbEndpoint0, USB_REQUEST_STATUS_STALLED);
		usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_SST);
		usbControlPipeSetState(USB_CP_WAIT_FOR_SETUP);
	}
	
	/* handle setup stage */
	if (udccsr0 & UDCCSR0_SA) {
		usbCPErr("Status Stage");
		usbEndpointNuke(&usbEndpoint0, USB_REQUEST_STATUS_SUCCESS);
		usbControlPipeSetState(USB_CP_SETUP_STAGE);
	}
	
	switch (usbUdc.cpState) {
	case USB_CP_WAIT_FOR_SETUP:
		/* hw bug */
		break;
	case USB_CP_SETUP_STAGE:
		if ((udccsr0 & UDCCSR0_CTRL_REQ) == UDCCSR0_CTRL_REQ) {
			readAndHandleControlRequest();
		}
		break;
	case USB_CP_IN_DATA_STAGE:
		if (usbEndpointHasPkt(&usbEndpoint0))
			usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_OPC);
		if (req && !usbEndpointIsFull(&usbEndpoint0))
			if (usbControlPipeWrite(req))
				usbControlInRequestComplete(req);
		break;
	case USB_CP_OUT_DATA_STAGE:
		if (usbEndpointHasPkt(&usbEndpoint0) && req)
			if (usbControlPipeRead(req))
				usbControlOutRequestComplete(req);
		break;
	case USB_CP_STALL:
		usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_FST);
		break;
	case USB_CP_IN_STATUS_STAGE:
		/* hw bug */
		if (pcInt)
			usbControlPipeSetState(USB_CP_WAIT_FOR_SETUP);
		break;
	case USB_CP_OUT_STATUS_STAGE:
	default:
		ASSERT(FALSE);/* TODO: better assertion */
		usbControlPipeSetState(USB_CP_WAIT_FOR_SETUP);
		break;
	}
}

/** \brief complete in request on endpoint zero
 * \param req request to be completed */
void usbControlInRequestComplete(struct UsbRequest * req) {
	ASSERT(req->ep == &usbEndpoint0);
	usbRequestComplete(&usbEndpoint0, req, USB_REQUEST_STATUS_SUCCESS);
}

/** \brief complete out request on endpoint zero
 * \param req request to be completed */
void usbControlOutRequestComplete(struct UsbRequest * req) {
	ASSERT(req->ep == &usbEndpoint0);
	usbControlPipeSetState(USB_CP_OUT_STATUS_STAGE);
	usbRequestComplete(&usbEndpoint0, req, USB_REQUEST_STATUS_SUCCESS);
	usbControlPipeSetState(USB_CP_WAIT_FOR_SETUP);
}

/** \brief write a request to endpoint zero
 * \param req request to be written
 * \return true iff complete request was sent */
bool usbControlPipeWrite(struct UsbRequest * req) {
	uint32_t count = usbEndpointWritePacket(&usbEndpoint0, req);
	bool is_short = (count < usbEndpoint0.epMaxPacketSize);
	
	if (is_short)
		usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_IPR);
	
	return is_short;
}

/** \brief read into a request from endpoint zero
 * \param req request to be read in
 * \return true iff a complete request was received */
bool usbControlPipeRead(struct UsbRequest * req) {
	while (usbEndpointHasPkt(&usbEndpoint0)) {
		uint32_t count = usbEndpointReadPacket(&usbEndpoint0, req);
		usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_OPC);
		
		bool is_short = (count < usbEndpoint0.epMaxPacketSize);
		
		if (is_short || usbEndpoint0.req->actual >= usbEndpoint0.req->length)
			return TRUE;
	}
	
	return FALSE;
}

/** \brief receive a control request and handle it */
static void readAndHandleControlRequest() {
	uint32_t i;
	union {
		uint32_t buf[2];
		struct DeviceRequest req;
	} u;
	
	/* dequeue all requests from ep 0 */
	usbEndpointNuke(&usbEndpoint0, USB_REQUEST_STATUS_PROTOCOL);
	if (usbEndpointHasPkt(&usbEndpoint0) &&
		usbEndpointBytesRemain(&usbEndpoint0) == 0) {
		usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_OPC);
	}
	
	/* read request */
	for (i = 0; i < 2; ++i) {
		if (usbEndpointIsEmpty(&usbEndpoint0)) {
			/* protocol stall */
			usbCPErr("Protocol stall");
			usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_FST | UDCCSR0_FTF);
			usbControlPipeSetState(USB_CP_STALL);
			return;
		}
		u.buf[i] = UDCDR0;
	}
	
	bool extraBytes = !usbEndpointIsEmpty(&usbEndpoint0);
	while (!usbEndpointIsEmpty(&usbEndpoint0))
		i = UDCDR0;
	
	if (extraBytes) {
		usbCPErr("Protocol stall: Extra bytes");
		usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_FST | UDCCSR0_FTF);
		usbControlPipeSetState(USB_CP_STALL);
		return;
	}
	
	if (u.req.bmRequestType & USB_ENDPOINT_IN)
		usbControlPipeSetState(USB_CP_IN_DATA_STAGE);
	else
		usbControlPipeSetState(USB_CP_OUT_DATA_STAGE);
	
	/* enter data stage */
	usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_SA | UDCCSR0_OPC);
	
	/* handle request */
	if (!handleControlRequest(&u.req)) {
		usbCPErr("Protocol stall: Could not handle request of type %d "
			"and request %d", u.req.bmRequestType, u.req.bRequest);
		usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_FST | UDCCSR0_FTF);
		usbControlPipeSetState(USB_CP_STALL);
	}
}

/** \brief handle a control request
 * \param req device request to be handled
 * \return true iff control request could be handled */
static bool handleControlRequest(const struct DeviceRequest * req) {
	if (USB_REQUEST_TYPE(req->bmRequestType) == USB_REQUEST_TYPE_STANDARD &&
		req->bRequest == USB_GET_DESCRIPTOR) {
		/* get descriptor of device */
		switch (req->descriptor.bType) {
			case USB_DESCRIPTOR_DEVICE:
				sendDeviceDescriptor(req->wLength);
				break;
			case USB_DESCRIPTOR_CONFIGURATION:
				sendConfigDescriptor(req->descriptor.bIndex, req->wLength);
				break;
			case USB_DESCRIPTOR_STRING:
				sendStringDescriptor(req->descriptor.bIndex, req->wLength);
				break;
			default:
				usbCPErr("Unknown Request: %d\n", req->descriptor.bType);
				return FALSE;
				break;
		}
		return TRUE;
	} else if (USB_REQUEST_TYPE(req->bmRequestType) ==
		USB_REQUEST_TYPE_STANDARD &&
		req->bRequest == USB_SET_CONFIGURATION) {
		/* handled by usbControlCCInterrupt */
		return TRUE;
	} else {
		/* cannot handle request */
		return FALSE;
	}
}

/** \brief send Configuration Descriptor
 * \param index Configuration index
 * \param wLength requested length of the descriptor */
void sendConfigDescriptor(uint8_t index, uint16_t wLength) {
	uint16_t length;
	uint8_t ifaceIdx, settingIdx, endpointIdx;
	uint8_t * buf;
	uint16_t bufOffset;
	struct ConfigurationDescriptor * configDesc;
	
	/* get the requested configuration */
	ASSERT(index <= usbUdc.config->numConfigurations);
	const struct UsbConfiguration * conf = usbUdc.config->configurations[index];
	
	/* get the length of the configuration with all its interfaces
		and endpoints */
	length = usbConfigGetConfigurationLength(conf);
	
	/* allocate memory */
	buf = (uint8_t*)dmalloc(length);
	
	configDesc = (struct ConfigurationDescriptor *)buf;
	/* next descriptor offset */
	bufOffset = sizeof(struct ConfigurationDescriptor);
	
	/* fill configuration descriptor */
	configDesc->bLength = sizeof(struct ConfigurationDescriptor);
	configDesc->bDescriptorType = USB_DESCRIPTOR_CONFIGURATION;
	configDesc->wTotalLength = length;
	configDesc->bNumInterfaces = conf->numInterfaces;
	configDesc->bConfigurationValue = index + 1;
	configDesc->iConfiguration = usbConfigGetStringIndex(conf->confDescription);
	configDesc->bmAttributes = USB_CONFIGURATION_BUS_POWERED_10
		| (conf->confSelfPowered ? USB_CONFIGURATION_BUS_POWERED_11 : 0)
		| (conf->confRemoteWakeup ? USB_CONFIGURATION_REMOTE_WAKEUP : 0);
	configDesc->bMaxPower = USB_POWER(conf->confMaxPower);
	
	/* iterate over all interfaces */
	for (ifaceIdx = 0; ifaceIdx < conf->numInterfaces; ++ifaceIdx) {
		/* get interface */
		const struct UsbInterface * iface = conf->interfaces[ifaceIdx];
		
		/* iterate over all alternate settings */
		for (settingIdx = 0; settingIdx < iface->numAltSettings; ++settingIdx) {
			const struct UsbAltSetting * setting = iface->altSettings[settingIdx];
			
			/* get interface descriptor */
			struct InterfaceDescriptor * ifaceDesc =
				(struct InterfaceDescriptor*)(buf + bufOffset);
			bufOffset += sizeof(struct InterfaceDescriptor);
			
			/* fill interface descriptor */
			ifaceDesc->bLength = sizeof(struct InterfaceDescriptor);
			ifaceDesc->bDescriptorType = USB_DESCRIPTOR_INTERFACE;
			ifaceDesc->bInterfaceNumber = ifaceIdx;
			ifaceDesc->bAlternateSetting = settingIdx;
			ifaceDesc->bNumEndpoints = setting->numEndpoints;
			ifaceDesc->bInterfaceClass = setting->altClass;
			ifaceDesc->bInterfaceSubClass = setting->altSubClass;
			ifaceDesc->bInterfaceProtocol = setting->altProtocol;
			ifaceDesc->iInterface =
				usbConfigGetStringIndex(setting->altDescription);
			
			/* iterate over all endpoints of the interface */
			for (endpointIdx = 0; endpointIdx < setting->numEndpoints;
				++endpointIdx) {
				/* get endpoint */
				const struct UsbEndpoint * endpoint = setting->endpoints[endpointIdx];
				
				/* get endpoint descriptor */
				struct EndpointDescriptor * endpointDesc =
					(struct EndpointDescriptor*)(buf + bufOffset);
				bufOffset += sizeof(struct EndpointDescriptor);
				
				/* fill endpoint descriptor */
				endpointDesc->bLength = sizeof(struct EndpointDescriptor);
				endpointDesc->bDescriptorType = USB_DESCRIPTOR_ENDPOINT;
				endpointDesc->bEndpointAddress =
					USB_ENDPOINT(endpoint->epAddress) |
					(endpoint->epDirection == USB_ENDPOINT_DIR_IN ?
						USB_ENDPOINT_IN : USB_ENDPOINT_OUT);
				endpointDesc->bmAttributes =
					endpoint->epType == USB_ENDPOINT_TYPE_CONTROL ?
						USB_ENDPOINT_CONTROL :
					(endpoint->epType == USB_ENDPOINT_TYPE_BULK ?
						USB_ENDPOINT_BULK :
					(endpoint->epType == USB_ENDPOINT_TYPE_INTERRUPT ?
						USB_ENDPOINT_INTERRUPT :
						USB_ENDPOINT_ISOCHRONOUS));
				endpointDesc->wMaxPacketSize =
					USB_ENDPOINT_PACKET_SIZE(endpoint->epMaxPacketSize);
				endpointDesc->bInterval = 0;
			}
		}
	}
	usbCPInfo("Sending Config Descriptor Request, index: %d "
		"with wLength = %d", index, wLength);
	
	/* send request, free memory */
	sendDescriptor(configDesc, length, wLength, TRUE);
}

/** \brief send String Descriptor
 * \param index String index
 * \param wLength requested length of the descriptor */
void sendStringDescriptor(uint8_t index, uint16_t wLength) {
	if (index == 0) {
		/* String index zero is special as it encodes the
			available languages only */
		static aligned4 struct StringDescriptor0 stringDesc = {
			/* one language available */
			.bLength = sizeof(stringDesc) + 1 * sizeof(*stringDesc.wLANGID),
			.bDescriptorType = USB_DESCRIPTOR_STRING,
			.wLANGID = { 0x0000 /* placeholder */ }
		};
		
		stringDesc.wLANGID[0] = usbUdc.config->languageID;
		
		usbCPInfo("Sending String 0 Descriptor, wLength = %d", wLength);
		
		/* send request, do not free memory */
		sendDescriptor(&stringDesc, stringDesc.bLength, wLength, FALSE);
	} else {
		const char * string;
		uint8_t len;
		uint8_t size, count;
		struct StringDescriptor * stringDesc;
		
		string = usbConfigGetString(index);
		if (string == NULL)
			len = 0;
		else
			len = strlen(string);
		/* String size is doubled as it is encoded unicode */
		size = sizeof(struct StringDescriptor) + 2 * len;
		
		/* allocate space */
		stringDesc = (struct StringDescriptor *)dmalloc(size);
		
		stringDesc->bLength = size;
		stringDesc->bDescriptorType = USB_DESCRIPTOR_STRING;
		/* copy string with simple unicode conversion */
		for (count = 0; count < len; ++count)
			stringDesc->bString[count] = string[count];
		
		usbCPInfo("Sending String Descriptor %d '%s', wLength = %d", index,
			string, wLength);
		
		/* send request, free memory after sending */
		sendDescriptor(stringDesc, size, wLength, TRUE);
	}
}

/** \brief send Device Descriptor
 * \param wLength requested length of the descriptor */
void sendDeviceDescriptor(uint16_t wLength) {
	/* the descriptor itself */
	static aligned4 struct DeviceDescriptor deviceDesc = {
		.bLength = sizeof(deviceDesc),
		.bDescriptorType = USB_DESCRIPTOR_DEVICE,
		.bcdUSB = 0x0110, /* usb 1.1 */
		.bMaxPacketSize0 = 16
	};
	
	const struct UsbDevice * devConfig = usbUdc.config;
	
	/* fill descriptor */
	deviceDesc.bDeviceClass = devConfig->devClass;
	deviceDesc.bDeviceSubClass = devConfig->devSubClass;
	deviceDesc.bDeviceProtocol = devConfig->devProtocol;
	deviceDesc.idVendor = devConfig->devIdVendor;
	deviceDesc.idProduct = devConfig->devIdProduct;
	deviceDesc.bcdDevice = devConfig->devBCDDevice;
	deviceDesc.iManufacturer =
		usbConfigGetStringIndex(devConfig->devManufacturer);
	deviceDesc.iProduct = usbConfigGetStringIndex(devConfig->devProduct);
	deviceDesc.iSerialNumber =
		usbConfigGetStringIndex(devConfig->devSerialNumber);
	deviceDesc.bNumConfigurations = devConfig->numConfigurations;
	
	usbCPInfo("Sending Device Descriptor, wLength = %d", wLength);
	
	/* send request, do not free memory */
	sendDescriptor(&deviceDesc, deviceDesc.bLength, wLength, FALSE);
}

/** \brief send descriptor on endpoint zero
 * \param descriptor descriptor to be sent
 * \param size size of descriptor
 * \param maxLength maximum length to be sent
 * \param freeData free data after sending is complete or cancelled
 * \see usbCommunicationSendDataIntr */
void sendDescriptor(const void * descriptor, uint8_t size,
	uint16_t maxLength, bool freeData) {
	ASSERT(usbEndpoint0.req == NULL);
	
	usbCPDebug("Sending Descriptor of length %d; Free: %s",
		min(size, maxLength), freeData ? "yes" : "no");
	
	usbRequestFill(&usbRequest0, &usbEndpoint0, (void*)descriptor,
		min(size, maxLength), freeData ? &cbFreeRequest : NULL);
	usbRequestEnqueue(&usbRequest0);
}

/** \brief callback of endpoint 0 requests which have data to be freed
 * \param req request */
static void cbFreeRequest(struct UsbRequest * req) {
	/* free data */
	usbCPDebug("Freeing data of last request");
	dfree(req->data);
}

/** @} */

/** \endcond */

