/* USB Driver: Main Interface
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <os_arm_bare_registers_def.h>
#include "../driver.h"
#include "usb_main.h"
#include "usb_main_internal.h"
#include "usb_configuration_internal.h"
#include "usb_endpoint_internal.h"
#include "usb_debug.h"

/** \defgroup usb Universal Serial Bus (USB) Subsystem
 * @{
 */

/** \file usb_main.c
 * \brief USB Initialization
 */

/** GPIO to recognize USB Power */
#define GPIO_USB_REC 13
/** GPIO to enable recognizing USB Power */
#define GPIO_USB_REC_EN 88

static void init(struct Driver * driver);
static void start(struct Driver * driver);
static void stop(struct Driver * driver);
static void finish(struct Driver * driver);

/** \brief USB Driver structure */
struct Driver usbDriver = {
	.name = "USB",
	
	.init = &init,
	.start = &start,
	.stop = &stop,
	.finish = &finish
};

/** \cond INTERNAL */

/** \brief global udc status structure */
struct UsbUdc usbUdc;

/** \brief Setup GPIOs for USB Recognition
 *
 * setup gpios to recognize wether we are powered by usb or battery */
static void init(struct Driver * driver) {
	/* initialize usb gpios for power recognition */
	GPIO_SET_ALT_FUNC(GPIO_USB_REC, 0, GPIO_IN);
	GPIO_SET_ALT_FUNC(GPIO_USB_REC_EN, 0, GPIO_OUT);
	SET_GPIO(GPIO_USB_REC_EN);
}

/** \brief configure and start USB Device Controller
 * \note configure subsystem first using usbConfigLoad() */
static void start(struct Driver * driver) {
	/* clear all interrupt sources */
	UDCISR0 = 0;
	UDCISR1 = 0;
	usbClear_UDCCR(UDCCR_UDE);
	
	/* enable usb clock */
	CKEN |= CKEN_CKEN11;
	
	/* reset internal usb configuration */
	usbUdc.configIdx = 0;
	usbUdc.ifaceIdx = 0;
	usbUdc.settingIdx = 0;
	usbUdc.cpState = USB_CP_WAIT_FOR_SETUP;
	usbUdc.ep[0] = &usbEndpoint0;
	usbUdc.endpointsUsed = 1;
	usbUdc.curAltSetting = NULL;
	
	/* initialize endpoint 0 */
	usbControlPipeInit();
	
	/* configure udc */
	usbConfigDeploy();
	
	/* enable udc */
	usbClear_UDCCR(UDCCR_OEN);
	usbSet_UDCCR(UDCCR_UDE);
	
	/* TODO: delay 2 us */
	if (UDCCR & UDCCR_EMCE) {
		usbErr("USB Configuration errors, udc disabled");
		/* TODO: handle that */
		return;
	}
	
	/* TODO msleep 100 */
	
	/* interrupt on configuration change and reset */
	UDCICR1 = UDCICR_IECC | UDCICR_IERS;
	
	UDCICR0 = UDC_INT(UDC_EP_0, UDCICR_IE);
	
	/* enable UDC Interrupt */
	ICMR_REG(IID_USBC) |= INTERRUPT_BIT(IID_USBC);
	
	usbInfo("started");
}

/** \brief stop USB Device Controller
 *
 * stops all USB activities and stops the controller */
static void stop(struct Driver * driver) {
	/* disable UDC Interrupt */
	ICMR_REG(IID_USBC) &= ~INTERRUPT_BIT(IID_USBC);
	
	/* clear all interrupt sources */
	UDCISR0 = 0;
	UDCISR1 = 0;
	
	/* ack pending interrupts */
	UDCISR0 |= 0;
	UDCISR1 |= 0;
	
	/* disable udc */
	usbClear_UDCCR(UDCCR_UDE);
	
	/* disable clock */
	CKEN &= ~CKEN_CKEN11;
	
	/* complete all requests */
	usbEndpointsFinishAll(USB_REQUEST_STATUS_SHUTDOWN);
}

/** \brief stop USB Subsystem */
static void finish(struct Driver * driver) {
	CLEAR_GPIO(GPIO_USB_REC_EN);
}

/** \endcond */

/** \brief check wether we're powered by usb or battery
 * \return \c TRUE iff imote is powered by usb */
bool usbBusPowered() {
	return GPLR(GPIO_USB_REC) & GPIO_BIT(GPIO_USB_REC) ? TRUE : FALSE;
}

/** @} */

